// ****************************************************************************
// BioVEC
//   A program for biomolecule visualization with ellipsoidal coarse-graining
// 
//
// Requires the OpenGL, FreeGlut, and DevIL libraries.
//
//
//                                                    Erik Abrahamsson, 2008
//
// Glacier:
//          gcc -Wall -I/usr/X11R6/include/ -o biovec.exe -L/usr/X11R6/lib -lX11 -lXi -lXmu -lglut -lGL -lGLU -lm BioVEC_Linux.cpp
// Jayz:
//          gcc -Wall -o biovec.exe -lglut -lGL -lGLU  -lIL -lILU -lILUT BioVEC_Linux.cpp 
//
//
// * License *******************************************************************
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License, version 3, as 
//    published by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/> 
//
//
// ****************************************************************************

#include <math.h>
#include <vector>
//#include <cstdlib> // C Standard General Utilities Library
#include <iostream>
#include <string>
//#include <stdio.h> // C library to perform Input/Output operations
//#include <stdlib.h> //C Standard General Utilities Library
#include <fstream>
#include <sstream>
//#include <GL/glut.h>
//#include <GL/freeglut.h>
#include <GL/openglut.h>
//#include <IL/il.h>
//#include <IL/ilu.h>
#include <IL/ilut.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_eigen.h>
//#include <gsl/gsl_blas.h>

using namespace std;

void Display();
void IdleDisplay();
void MyGLInit();
void MyInit(int argc, char *argv[]);
void CreateEllipsoid(int, vector<double>, vector<double>, int, int);
void DoStuff();
void DrawLineBonds(int, vector<double>);
void DrawSolidBonds(int, vector<double>);
string ReadConfig(string);
void ReadBonds(string);
void OpenFile();
void OpenQuatFile();
void OpenAnisouFile();
void CloseFile();
void ReadFrame();
void ReadQuatFrame();
void ReadAnisouFrame();
void PrintImage(bool);

void createGLUTMenus();
void MenuHandler(int);
void Reshape(int, int);
void KeyHandler(unsigned char, int, int);
void SpecialKey(int, int, int);
void getMVCoord(double[], double[], double[]);
double dot(double[], double[]);
vector<double> average();

void ShowTimeStep();
void resetPerspectiveProjection();
void setOrthographicProjection();
void renderBitmapString(double x, double y, void *bmfont, char *instring);

void LoadTextures(string, int);
void DeleteTextures();

void HandleMouseState(int button, int state, int x, int y);
void HandleMouseMove(int x, int y);
void PrintMouseState();
void HandleWheel(int, int, int, int);


double pi = 3.14159265358979323846;

int *bitfont = (int *) GLUT_BITMAP_8_BY_13;
int bitmapHeight = 13;
int rotating_light = 1; // Should the light be stationary or not?
int window_width = 1024,
    window_height = 640,
    window_depth = 400;
int atoms = 0; // Number of atoms in file
int species = 0; // Number of different species
int frame = 0; // Frame count
int printcount = 0; // Printing counter

double bondwidth = 0; // Width of soldi bonds

// Half radii of ellipsoids and radii of spheres, and colors used
vector<double> shape,
               bondvec,
               color;
vector<double> geom; // Vector containing all geometries

vector<int> counter(100, 0); // Frame counter
vector<int> topology;        // 
vector<GLuint> texture(1, 0);

bool animate = false;     // Animate?
bool print = false;       // Print images to file?
bool showTS = true;       // Show timestep?
bool gp = true;			  // Fog or not?
bool solidBonds = false;  // Draw bonds as cylinders?
bool quat = false;

ifstream infile; // Dump file stream

string inname;   // Name of dump file
string format = ".jpg"; 
string ellipsoidtexture, bondtexture, title;

float FogColor[4];// = {0.08f, 0.09f, 0.19f, 1.0f}; // {0.5, 0.5, 0.5, 0.0}; // 
float BGColor[4];// = {0.08, 0.09, 0.19, 0.0};

float light_pos_0[] = {50, 0, 100, 0},
       light_pos_1[] = {50, 50, -100, 0};

float light_specular[] = {0.0f, 0.0f, 0.0f, 1.0f}; //  sets specular highlight of balls
float light_diffuse[] = {0.9f, 0.9f, 0.9f, 1.0f};
float light_ambient[]  = {0.1f, 0.1f, 0.1f ,1.0f};  // global ambient
float light_ambient_2[] = {0.1f, 0.1f, 0.1f ,1.0f}; // ambient of lightsource

GLUquadricObj	*q;

double eyex, eyey, eyez;  // eye x,y,z values for gluLookAt (location of eye)
double  focusx, focusy, focusz; // the point the eye is looking at
double g_xRotation, g_yRotation;

struct MouseState {
	bool leftButton;
	bool rightButton;
	bool middleButton;
	int func_key;
	int mx;
	int my;
} MouseState;


/* MAIN ******************************************************************** */
int main(int argc, char *argv[]) {
   
   glutInit(&argc, argv);
   // Create window
   // GLUT_DOUBLE: Double buffering
   glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
   glutInitWindowSize(window_width, window_height);
   
   MyInit(argc, argv);       // My own initializations
   
   glutCreateWindow(title.c_str());
   
   // Draw the scene
   glutDisplayFunc(&Display);
   // Idle - draw!
   glutIdleFunc(&Display);
   // Reshape window
   glutReshapeFunc(&Reshape); 
   glutKeyboardFunc(&KeyHandler);
   glutSpecialFunc(&SpecialKey);
   // Mouse events
   glutMouseFunc(HandleMouseState);
   glutMotionFunc(HandleMouseMove);
   glutMouseWheelFunc(HandleWheel);
   // Right click menu
   createGLUTMenus(); // Create menu
   
   MyGLInit();
   
   glutMainLoop(); // Wait for events
   
   // Close the dump file
   CloseFile();  
   
   // Delete the textures
   DeleteTextures();
   
   return 0;       // End main
}


/* DISPLAY ***************************************************************** */
void Display() {

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // clear buffers
   
   DoStuff();
   
   if (animate) {
      if (showTS) 
         ShowTimeStep();
      if (print) 
         PrintImage(false);
      if (frame < (int) counter.size()-1) {
         frame++;
         ReadFrame();
      }
    }
    else {
       if (showTS)
          ShowTimeStep();
    }
    
    glutSwapBuffers(); // double buffering
}


/* IDLEDISPLAY ************************************************************* */
void IdleDisplay() {

   if (animate)
      Display();
}


/* MYGLINIT **************************************************************** */
void MyGLInit() {

   float df = 0.0;
   
   glClearColor(BGColor[0], BGColor[1], BGColor[2], BGColor[3]);
   
   glEnable(GL_TEXTURE_2D);
   
   glShadeModel(GL_SMOOTH);							// Enable Smooth Shading
   glClearDepth(1.0f);									  // Depth Buffer Setup
   glEnable(GL_DEPTH_TEST);							// Enables Depth Testing
   glDepthFunc(GL_LEQUAL);								// The Type Of Depth Testing To Do
   glEnable(GL_CULL_FACE);               // Don't draw the back of polygons
   glCullFace(GL_BACK);                  // saves time!
   glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	// Really Nice Perspective Calculations
   glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
   glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
   glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
   
   glMaterialfv(GL_FRONT, GL_SPECULAR, light_specular);
   glMaterialfv(GL_FRONT, GL_SHININESS, &df);
   
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(45.0f, (GLdouble)window_width/(GLdouble)window_height, 1, window_depth);
   
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity(); // Matrix = I
   
   eyex = 0;
   eyey = 0;
   eyez = 200;
   focusx = focusy = focusz = 0.0f;
   
   //gluLookAt(0,0,200, 0,0,0, 0,1,0); // eye position, center, direction of up
   gluLookAt(eyex, eyey, eyez, focusx, focusy, focusz, 0,1,0);
   
   glFogi(GL_FOG_MODE, GL_LINEAR);  // Fog Mode
   glFogfv(GL_FOG_COLOR, FogColor);          // Set Fog Color
   glFogf(GL_FOG_DENSITY, 100);            // How Dense Will The Fog Be
   glHint(GL_FOG_HINT, GL_DONT_CARE);        // Fog Hint Value
   glFogf(GL_FOG_START, (GLdouble)window_depth - 50);               // Fog Start Depth
   glFogf(GL_FOG_END, (GLdouble)window_depth);                 // Fog End Depth
   glEnable(GL_FOG);                         // Enables GL_FOG
   
   // set up light 0
   glEnable(GL_LIGHTING);
   glLightfv(GL_LIGHT0, GL_AMBIENT,  light_ambient_2); 
   glLightfv(GL_LIGHT0, GL_DIFFUSE,  light_diffuse); 
   glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular); 
   glEnable(GL_LIGHT0);
   
   glLightfv(GL_LIGHT0, GL_POSITION, light_pos_0); // place light
   
   // set up light 1
   glEnable(GL_LIGHTING);
   glLightfv(GL_LIGHT1, GL_AMBIENT,  light_ambient_2); 
   glLightfv(GL_LIGHT1, GL_DIFFUSE,  light_diffuse); 
   glLightfv(GL_LIGHT1, GL_SPECULAR, light_specular); 
   glEnable(GL_LIGHT1); 
   glLightfv(GL_LIGHT1, GL_POSITION, light_pos_1); // place light
   
   q = gluNewQuadric();								// Create A New Quadratic
   gluQuadricNormals(q, (GLenum) GLU_SMOOTH);					// Generate Smooth Normals For The Quad
   gluQuadricTexture(q, GL_TRUE);						// Enable Texture Coords For The Quad
   
   q = gluNewQuadric();							// Create A New Quadratic
   
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
   
   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE); // Inverse texture and color
   
   ilInit();
   iluInit();
   ilutInit();
   ilutRenderer(ILUT_OPENGL);
   
   if (ellipsoidtexture.length() != 0)
      LoadTextures(ellipsoidtexture, 0);
   if (bondtexture.length() != 0)
      LoadTextures(bondtexture, 1);
   
}


/* MYINIT ****************************************************************** */
void MyInit(int argc, char *argv[]) {

   string bondfile, conffile;
   
   if ( argc != 2 ) {
      cout << "Please enter name of file with configuration information: " ;
      getline (cin, conffile);
      title = "BioVEC - " + conffile; 
   }
   else {
      conffile = argv[1];
      title = "BioVEC - ";
      title += argv[1];
   }
   bondfile = ReadConfig(conffile);
   cout << "Config file read." << endl; 
   
   OpenFile();
   ReadFrame();
   
   topology.clear();
   if (!bondfile.empty()) {
      ReadBonds(bondfile);
   }
   
   cout << "Done reading. Rendering." << endl;
   
}


/* READCONFIG ************************************************************** */
string ReadConfig(string fname) {

   ifstream conffile;
   string bondfile, text, st;
   int i, id = 0, k, m, it, eofile = 0;
   double a, b, c;
   vector<string> keyword(11," ");
   vector<int> set(keyword.size(),0);
   
   keyword.at(0) = "dumpfile";
   keyword.at(1) = "bondfile";
   keyword.at(2) = "species";
   keyword.at(3) = "half_radii";
   keyword.at(4) = "bond_vector";
   keyword.at(5) = "color";
   keyword.at(6) = "background";
   keyword.at(7) = "solid_bonds";
   keyword.at(8) = "texture";
   keyword.at(9) = "texture_bonds";
   keyword.at(10) = "anisoufile";
   
   conffile.open(fname.c_str());
   
   if (conffile.is_open()) {
      cout << "Input file " << fname << " is opened." << endl; 
      
      // Clear everytning
      shape.clear();
      bondvec.clear();
      color.clear();
      
      
      while (eofile == 0) {
         // Read lines until end of file
         getline(conffile, st);
         //cout << "st: *" << st << "*" << endl;
         
         // If the line is blank, ignore and clear st
         i = 0;
         for (int j = 0; j < (int) st.size(); j++) {
            if (!isspace(st.at(j))) {
               i = 1;
               break;
            }
         }
         if (i == 0)
            st.clear();
            
         // Check if not a comment  
         if (!st.empty()) {
            text = st.at(0);
            if (text.compare("#") != 0) {
               // Clear everything
               i = 0;
               text.clear();
               while (i < (int) st.length() && !isspace(st.at(i))) {
                  text += st.at(i);
                  i++;
               }
               
               for (int j = 0; j < (int) keyword.size(); j++) {
                  if (text.compare(keyword.at(j)) == 0) {
                     id = j;
                     //cout << id << endl;
                     break;
                  }
                  else
                     id = keyword.size();
                     //cout << id << endl;
               }
                           
               switch(id) {
                  
                  case 0:
                     // Found dumpfile - the file with simupation data
                     //cout << "Found " << text << endl;
                     if (set.at(10) == 0) {
                        inname.clear();
                        // Find next word
                        while (i < (int) st.length() && isspace(st.at(i)))
                           i++;
                        while (i < (int) st.length() && !isspace(st.at(i))) {
                           inname += st.at(i);
                           i++;
                        }
                        cout << "Dump File: " << inname << endl;
                        set.at(id) = 1;
                        set.at(10) = 1; // No need to read anisou data.
                        quat = true;
                        break;
                     }
                     else {
                        cout << "Anisou file already set. Which one do you want?" << endl;
                        cout << "Please correct this, and run the program again." << endl;
                        cout << "Press Enter to exit.";
                        cin.clear();
                        cin.ignore(1,0);
                        std::exit(0); 
                     }
                  case 1:
                     // Found bondfile - the file with bond data
                     //cout << "Found " << text << endl;
                     bondfile.clear();
                     // Find next word
                     while (i < (int) st.length() && isspace(st.at(i)))
                        i++;
                     while (i < (int) st.length() && !isspace(st.at(i))) {
                        bondfile += st.at(i);
                        i++;
                     }
                     cout << "Bond File: " << bondfile << endl;
                     set.at(id) = 1;
                     break;
                  
                  case 2:
                     // Found species - the number of species
                     //cout << "Found " << text << endl;
                     text.clear();
                     // Find next word
                     while (i < (int) st.length() && isspace(st.at(i)))
                        i++;
                     while (i < (int) st.length() && !isspace(st.at(i)) && isdigit(st.at(i))) {
                        text += st.at(i);
                        i++;
                     }
                     istringstream(text) >> species;           
                     cout << "The number of species is " << species << endl;
                     shape.resize(species*3);
                     bondvec.resize(species*3);
                     color.resize(species*3);
                     set.at(id) = 1;
                     break;
                  
                  case 3:
                     // Found half_radii - the half-radii of the species
                     //cout << "Found " << text << endl;
                     if (set[2] == 1) {
                        text.clear();
                        while (i < (int) st.length()) {
                           text += st.at(i);
                           i++;
                        }
                        //cout << text << endl;
                        a = 0;
                        b = 0;
                        c = 0;
                        istringstream(text) >> k >> a >> b >> c; //shape.at((k-1)*3+0) >> shape.at((k-1)*3+1) >> shape.at((k-1)*3+2);
                        shape.at((k-1)*3+0) = a;
                        shape.at((k-1)*3+1) = b;
                        shape.at((k-1)*3+2) = c;
                        set.at(id) = 1;
                     }
                     else {
                        cout << "The number of species is not set yet."<< endl;
                        cout << "This has to be done before the half radii are read." << endl;
                        cout << "Please move the number of species above the half radii in the input file" << endl;
                        cout << "and run the program again." << endl;
                        cout << "Press Enter to exit.";
                        cin.clear();
                        cin.ignore(1,0);
                        std::exit(0);
                     }
                     break;
                  
                  case 4:
                     // Found bond_vectors - the bond vectors
                     //cout << "Found " << text << endl;
                     if (set[2] == 1) {
                        text.clear();
                        while (i < (int) st.length()) {
                           text += st.at(i);
                           i++;
                        }
                        //cout << text << endl;
                        a = 0;
                        b = 0;
                        c = 0;
                        istringstream(text) >> k >> m >> a >> b >> c; //shape.at((k-1)*3+0) >> shape.at((k-1)*3+1) >> shape.at((k-1)*3+2);
                        
                        it = (m-1)*species*3 + (k-1)*3;
                        //cout << it << " " << bondvec.size() << endl;
                        if (it > (int) bondvec.size()-1) {
                           //cout << "Resize" << endl; 
                           bondvec.resize(it + 3);
                        }
                        bondvec.at(it + 0) = a;
                        bondvec.at(it + 1) = b;
                        bondvec.at(it + 2) = c;
                        
                        set.at(id) = 1;
                     }
                     else {
                        cout << "The number of species is not set yet."<< endl;
                        cout << "This has to be done before the bond vectors are read." << endl;
                        cout << "Please move the number of species above the half radii in the input file" << endl;
                        cout << "and run the program again." << endl;
                        cout << "Press Enter to exit.";
                        cin.clear();
                        cin.ignore(1,0);
                        std::exit(0);
                     }
                     break;
                  
                  case 5:
                     // Found colors or colours - the colours of the species
                     //cout << "Found " << text << endl;
                     if (set[2] == 1) {
                        text.clear();
                        while (i < (int) st.length()) {
                           text += st.at(i);
                           i++;
                        }
                        a = 0;
                        b = 0;
                        c = 0;
                        //cout << text << endl;
                        istringstream(text) >> k >> a >> b >> c; //shape.at((k-1)*3+0) >> shape.at((k-1)*3+1) >> shape.at((k-1)*3+2);
                        color.at((k-1)*3+0) = a;
                        color.at((k-1)*3+1) = b;
                        color.at((k-1)*3+2) = c;
                     }
                     else {
                        cout << "The number of species is not set yet."<< endl;
                        cout << "This has to be done before the color vectors are read." << endl;
                        cout << "Please move the number of species above the colors in the input file" << endl;
                        cout << "and run the program again." << endl;
                        cout << "Press Enter to exit.";
                        cin.clear();
                        cin.ignore(1,0);
                        std::exit(0);
                     }
                     set.at(id) = 1;
                     break;
                  
                  case 6:
                     // Found background - background colour
                     //cout << "Found " << text << endl;
                     text.clear();
                     while (i < (int) st.length()) {
                        text += st.at(i);
                        i++;
                     }
                     istringstream(text) >> BGColor[0] >> BGColor[1] >> BGColor[2] >> st;
                     BGColor[3] = 1.0;
                     for (int j = 0; j < 3; j++) {
                        FogColor[j] = BGColor[j];
                     }
                     set.at(id) = 1;
                     
                     break;
                  
                  case 7:
                     // Found solid_bonds
                     text.clear();
                     while (i < (int) st.length()) {
                        text += st.at(i);
                        i++;
                     }
                     istringstream(text) >> bondwidth;
                     if (bondwidth > 0.0) {
                        solidBonds = true;
                        set.at(id) = 1;
                        //cout << bondwidth << endl;
                     }
                     break;
                     
                  case 8:
                     // Found texture - name of the texture file
                     //cout << "Found " << text << endl;
                     // Make sure we have room in texture vector
                     if (texture.empty())
                        texture.resize(1);
                     //Clear file name 
                     ellipsoidtexture.clear();
                     // Find next word
                     while (i < (int) st.length() && isspace(st.at(i)))
                        i++;
                     while (i < (int) st.length() && !isspace(st.at(i))) {
                        ellipsoidtexture += st.at(i);
                        i++;
                     }
                     set.at(id) = 1;
                     break;
                  
                  case 9:
                     // Found texture - name of the texture file
                     //cout << "Found " << text << endl;
                     // Make sure we have room in texture vector
                     if (texture.size() < 2)
                        texture.resize(2);
                     //Clear file name 
                     bondtexture.clear();
                     // Find next word
                     while (i < (int) st.length() && isspace(st.at(i)))
                        i++;
                     while (i < (int) st.length() && !isspace(st.at(i))) {
                        bondtexture += st.at(i);
                        i++;
                     }
                     //cout << texturefile << endl;
                     
                     set.at(id) = 1;
                     break;
                  
                  case 10:
                     // Found anisoufile - the file with simupation data
                     //cout << "Found " << text << endl;
                     if (set.at(0) == 0) {
                        inname.clear();
                        // Find next word
                        while (i < (int) st.length() && isspace(st.at(i)))
                           i++;
                        while (i < (int) st.length() && !isspace(st.at(i))) {
                           inname += st.at(i);
                           i++;
                        }
                        cout << "ANISOU File: " << inname << endl;
                        set.at(id) = 1;
                        set.at(0) = 1; // Can't have both anisou and quat
                        set.at(2) = 1; // no need to set number of species
                        set.at(3) = 1; // no need to set shapes
                        set.at(4) = 1; // no need to set bondvectors
                        set.at(5) = 1; // no need to set colors
                        species = 1;
                        quat = false;
                        break;
                     }
                     else {
                        cout << "Quat file already set. Which one do you want?" << endl;
                        cout << "Please correct this, and run the program again." << endl;
                        cout << "Press Enter to exit.";
                        cin.clear();
                        cin.ignore(1,0);
                        std::exit(0); 
                     }
                     
                  default:
                     cout << "Wrong keyword at the entry '" << st << "'." << endl;
                     cout << "Please correct the input file, and run the program again." << endl; 
                     cout << "Press Enter to exit.";
                     cin.clear();
                     cin.ignore(1,0);
                     std::exit(0);
                     break;
               }
            }
         }
         
         // End of file found
         if (conffile.eof())
            eofile = 1;
      }
      
      // Check size of bondvec. Needs to be (bond origin) * species * 3.
      if ( fmod(bondvec.size(), 3*species) != 0) {
         //cout << bondvec.size() << " Mod " << fmod(bondvec.size(), 3*species) << endl;
         // This is a bit complicated to explain... Each "row" in the matrix has 3*species entries... But it's a vector...
         c = 3*species - fmod(bondvec.size(), 3*species) + bondvec.size();
         //cout << "c: " << c << endl;
         bondvec.resize((int) c);
      }
      
      // Check the optional keywords
      // If no keyword for bond file was found
      if (set.at(1) == 0) {
         cout << "No file with bond information found." << endl;
         cout << "No bonds will be drawn." << endl;
         set.at(1) = 1;
      }
      
      // If no keyword for solid bonds was found
      if (set.at(7) == 0) {
         //cout << "Bonds will be drawn as lines." << endl;
         solidBonds = false;
         set.at(7) = 1;
      }
      
      // If no keyword for texture was found
      if (set.at(8) == 0) {
         cout << "No texture will be used for the species." << endl;
         set.at(8) = 1;
      }
      
      // If no keyword for bond texture was found
      if (set.at(9) == 0) {
         //cout << "No texture will be used." << endl;
         set.at(9) = 1;
      }
      
      for (int j = 0; j < (int) set.size(); j++) {
         if (set.at(j) == 0) {
            cout << "The value of the " << keyword.at(j) << " is missing in the input." << endl;
            cout << "Please add this to the input file (see the manual) and run the program again." << endl;
            cout << "Press Enter to exit.";
            cin.clear();
            cin.ignore(1,0);
            std::exit(0);
         }
      }

      conffile.close();
      
   }
   
   else {
      cout << "The file " << fname << " could not be found." << endl;
      cout << "Please supply a configuration filed named containing" << endl;
      cout << "information on half radii, bond vectors and colours:  (Enter to exit) ";
      getline (cin, fname);
      title = "Biovec - " + fname;
      if (fname == "") {
         std::exit(0);
      }
      // Try ReadConfig() again
      bondfile = ReadConfig(fname);
   }
   
   return bondfile;
}


/* OPENFILE **************************************************************** */
void OpenFile() {
   
   if (quat) {
      OpenQuatFile();
   }
   else {
      OpenAnisouFile();
   }
}


/* OPENQUATFILE ************************************************************ */
void OpenQuatFile() {

   int count = 1, stepatoms;
   string text, cstring;
   
   // Open LAMMPS output file
   infile.open(inname.c_str(), ios::in | ios::binary);
   
   if (infile.is_open()) {
      cout << "Infile " << inname << " opened" << endl;
      cout << "Reading...";
      counter.resize(1);
      counter.at(0) = infile.tellg();
      
      // Read complete file to get number of timesteps
      // Read first step, to get number of atoms
      while (cstring != "ITEM: NUMBER OF ATOMS" && !infile.eof()) {
         getline(infile, text);
         cstring.assign(text, 0, 21);
         //cout << cstring << endl;
      }
      getline (infile, text);
      istringstream(text) >> atoms;
      stepatoms = atoms;
      while (cstring != "ITEM: ATOMS" && !infile.eof()) {
         getline(infile, text);
         cstring.assign(text, 0, 11);
      }
      for (int i = 0; i < atoms; i++) {
         getline (infile, text);
      }
           
      // First step read
      // Read the rest
      while (!infile.eof()) {
         counter.push_back(infile.tellg());
         while (cstring != "ITEM: NUMBER OF ATOMS" && !infile.eof()) {
            getline(infile, text);
            cstring.assign(text, 0, 21);
            //cout << cstring << endl;
         }
         getline (infile, text);
         istringstream(text) >> stepatoms;
         //cout << "stepatoms: " << stepatoms << endl;
         if (stepatoms != atoms) {
            cout << "Incorrect number of atoms in time step " << frame << "." << endl;
            cout << "Number of atoms read: " << stepatoms << endl;
            cout << "Please correct this and run program again." << endl;
            cout << "Press Enter to exit.";
            cin.clear();
            cin.ignore(1,0);
            std::exit(0);
         }
         while (cstring != "ITEM: ATOMS" && !infile.eof()) {
            getline(infile, text);
            cstring.assign(text, 0, 11);
         }
         for (int i = 0; i < atoms; i++) {
            getline (infile, text);
         }
         count++;
         if (count%100 == 0) cout << ".";
         if (count%5000 == 0) cout << endl << "This could take a while..." << endl;
      }
      cout << endl;
      if (text == "")
         counter.resize(count-1); // One line too much, due to blank line at end of file
         
      infile.clear();
      infile.seekg(0, ios::beg); // Rewind file
      
      // Allocate vector geom
      geom.resize(atoms*9);
      // cout << "Resizing geom. New size " << geom.size() << " elements." << endl;
      
      cout << "Number of atoms:      " << atoms << endl;
      cout << "Number of time steps: " << counter.size() << " (0 - " << counter.size()-1 << ")"<< endl;
      if ((int) counter.size() == 1 )
         showTS = false;
   }
   else {
      cout << "Error. Dump file " << inname << " not found." << endl;
      cout << "Please enter the correct name of the dump file in the input file. " << endl;
      cout << "Press Enter to exit.";
      cin.clear();
      cin.ignore(1,0);
      std::exit(0);
   }
}


/* OPENANISOUFILE*********************************************************** */
void OpenAnisouFile() {
   
   string cline, cstring;
   int stepatoms = 0, step = 0;
   
   atoms = 0;
   infile.open(inname.c_str(), ios::in | ios::binary);
   
   counter.clear();
   
   if (infile.is_open()) {
      cout << "Infile open" << endl;
      getline(infile, cline);
      cstring.assign(cline, 0, 4);
      while (!infile.eof()) {
         while (cstring != "CRYS" && !infile.eof()) {
            getline(infile, cline);
            cstring.assign(cline, 0, 4);
         }
         if (cstring == "CRYS") {
            //cout << "Found " << cstring << "!" << endl;
            counter.push_back(infile.tellg()); 
            step++;
            //cout << "step: " << step << endl;
            cstring.clear();
            while (cstring != "CRYS" && !infile.eof()) {
               getline(infile, cline);
               cstring.assign(cline, 0, 4);
               if (cstring == "ATOM" || cstring == "HETA") {
                  stepatoms++;
               }
            }
            //cout << atoms << endl;
            if (atoms == 0) {
               // This is the first frame
               atoms = stepatoms;
            }
            if (stepatoms != atoms) {
               cout << "Incorrect number of atoms in time step " << step << "." << endl;
               cout << "Number of atoms read: " << stepatoms << endl;
               cout << "Please correct this and run program again." << endl;
               cout << "Press Enter to exit.";
               cin.clear();
               cin.ignore(1,0);
               std::exit(0);
            }
            
            stepatoms = 0;
         }
         
         //getline(infile, cline);
         //cstring.assign(cline, 0, 4);
      }
      
      if (atoms == 0) {
         cout << "No keyword 'CRYS' found. Are you sure " << inname << " is a PDB file?" << endl;
         cout << "If it is, please add the keyword 'CRYS' at the begining of the file," << endl;
         cout << "and before each timestep." << endl;
         cout << "Press Enter to exit.";
         cin.clear();
         cin.ignore(1,0);
         std::exit(0);
      }
      
      infile.clear();
      infile.seekg(0, ios::beg); // Rewind file
      
      species = atoms;
      cout << "Number of atoms:      " << atoms << endl;
      cout << "Number of time steps: " << counter.size() << " (0 - " << counter.size()-1 << ")"<< endl;
      if ((int) counter.size() == 1 )
         showTS = false;
   }
   else {
      cout << "Error. Dump file " << inname << " not found." << endl;
      cout << "Please enter the correct name of the dump file in the input file. " << endl;
      cout << "Press Enter to exit.";
      cin.clear();
      cin.ignore(1,0);
      std::exit(0);
   }
}


/* CLOSEFILE *************************************************************** */
void CloseFile() {

   if (infile.is_open()) {
      infile.close();
   }
   else {
      cout << "Something is seriously wrong. The file is already closed." << endl;
      cout << "But we were going to quit anyway, so that's ok. For now." << endl;
      cout << "Press Enter to exit.";
      cin.clear();
      cin.ignore(1,0);
      std::exit(0);
   }
}


/* READBONDS *************************************************************** */
void ReadBonds(string fname) {

   ifstream file;
   string text;
   int a, b, c, d, linecount = 0;
   double e, f;
   
   topology.clear();
   
   file.open(fname.c_str(), ios::in | ios::binary);
   if (file.is_open()) {
      cout << "Bond information file " << fname << " is opened." << endl;
      //getline(file, text); // First line is junk 
      while(!file.eof()) { 
         getline(file, text);
         linecount++;
         //cout << text << endl;
         if (text.compare("Bonds")) {
            //getline(file, text);
            linecount++;
            //cout << text << endl;
            while (getline(file, text)) {
               linecount++;
               //cout << text << endl;
               e = -1; f = -1;
               istringstream(text) >> a >> b >> c >> d >> e >> f;
               if (isdigit(text.at(0))) {
                  //topology.push_back(b);
                  topology.push_back(c);
                  topology.push_back(d);
                  if (((int) e > 0) && ((int) f > 0)) {
                     topology.push_back((int) e);
                     topology.push_back((int) f);
                     }
                  else {
                     topology.push_back(1);
                     topology.push_back(1);
                  }
                  if ((c > atoms) || (d > atoms)) {
                     cout << "Error reading bonds. Bond goes to atom that does not exist." << endl;
                     cout << "Line number " << linecount << ": " << text << endl;
                     cout << "Please correct this and run program again." << endl;
                     cout << "Press Enter to exit.";
                     cin.clear();
                     cin.ignore(1,0);
                     std::exit(0);
                  }
                  if ( (e*species > bondvec.size()) || (f*species > bondvec.size())) {
                     cout << "Error reading bonds. Bond goes from bond origin that does not exist." << endl;
                     cout << "Line number " << linecount << ": " << text << endl;
                     cout << "Please correct this and run program again." << endl;
                     cout << "Press Enter to exit.";
                     cin.clear();
                     cin.ignore(1,0);
                     std::exit(0);
                  }
               }
            }
         }
      }
      file.close();
      //topology.resize(topology.size()-2); // always one line too much read
      cout << "Number of bonds:      " << topology.size()/4 << endl;
   }
   else {
      cout << "No file with bond information found." << endl;
      cout << "No bonds will be drawn. Reload input file (Shift-L) to load bond file." << endl;
   }
}


/* READFRAME *************************************************************** */
void ReadFrame() {
   
   if (quat) {
      ReadQuatFrame();
   }
   else {
      ReadAnisouFrame();
   }
}


/* READQUATFRAME *********************************************************** */
void ReadQuatFrame() {

   string text, cstring;
   //char c;
   //int stepatoms;
   
   if (infile.is_open()) {
      infile.seekg(counter.at(frame), ios::beg);
      
      // Dump file structure:
      //   Description of data
      //   <Data>
       
      // Find number of atoms in timestep
      //while (cstring != "ITEM: NUMBER OF ATOMS" && !infile.eof()) {
      //   getline(infile, text);
      //   cstring.assign(text, 0, 21);
      //   //cout << cstring << endl;
      //}
      //getline (infile, text);
      //istringstream(text) >> stepatoms;
      
      //// cout << "Number of atoms = " << atoms << endl;
      //if (stepatoms != atoms) {
      //   cout << "Incorrect number of atoms in time step " << frame << "." << endl;
      //   cout << "Number of atoms read: " << stepatoms << endl;
      //   cout << "Please correct this and run program again." << endl;
      //   cout << "Press Enter to exit.";
      //   cin.clear();
      //   cin.ignore(1,0);
      //   std::exit(0);
      //}
      
      while (cstring != "ITEM: ATOMS" && !infile.eof()) {
         getline(infile, text);
         cstring.assign(text, 0, 11);
      }
      // Read geometries
      for (int i = 0; i < atoms ; i++) {
         getline (infile, text);
         istringstream(text) >> geom.at(i*9+0) >> geom.at(i*9+1) >> 
	           geom.at(i*9+2) >> geom.at(i*9+3) >> geom.at(i*9+4) >> geom.at(i*9+5) >> 
	           geom.at(i*9+6) >> geom.at(i*9+7) >> geom.at(i*9+8);
         if (geom.at(i*9+1) > species) {
            cout << "Illegal entry for atom no " << geom.at(i*9) << " in time step " << frame << ":" << endl;
            cout << text << endl;
            cout << "Type read: " << geom.at(i*9+1) << ". Only " << species << " specified in input." << endl;
            cout << "Please correct this and run program again." << endl;
            cout << "Press Enter to exit.";
            cin.clear();
            cin.ignore(1,0);
            std::exit(0);
         }
      }
      //}
      //else {
         // It seems that we have reached the end of the file. 
         // Stop the animation, and step back to the previous frame, which was the last one.
      //   animate = false;
      //   frame--;
      //}
   }
   else {
      cout << "Something is seriously wrong. The input file seems to be closed." << endl;
      cout << "Press Enter to exit.";
      cin.clear();
      cin.ignore(1,0);
      std::exit(0);
   }
}


/* READANISOUFRAME ********************************************************* */
void ReadAnisouFrame() {
   
   gsl_matrix *A, *evec, *mtemp;
   gsl_eigen_symmv_workspace *w;
   gsl_vector *eval, *vec0, *vec1, *vec2, *vtemp;
   int info, count = 0, n = 3, atno; //, a0, a1; 
   string cline, cstring, pdb, psf, dump;
   char cjunk[80];
   //char type;
   vector<double> coord(3, 0);
   vector<double> aniso(6, 0);
   vector<char> type(200,' ');
   double trcP, W, X, Y, Z, S, norm;
   double m00, m01, m02, m10, m11, m12, m20, m21, m22;
   
   //ifstream infile;
   
   if (infile.is_open()) {
      //cout << "PDB ANISOU data is being read." << endl;
      //cout << "Let's go!" << endl;
      
      // Allocate and set to zero
      A = gsl_matrix_calloc (n, n);
      evec = gsl_matrix_calloc (n, n);
      mtemp = gsl_matrix_calloc (n, n);
      eval = gsl_vector_calloc (n);
      vec0 = gsl_vector_calloc (n);
      vec1 = gsl_vector_calloc (n);
      vec2 = gsl_vector_calloc (n);
      vtemp = gsl_vector_calloc (n);
      w = gsl_eigen_symmv_alloc(n);
      
      geom.clear();
      shape.clear();
      color.clear();
      bondvec.clear();
      infile.clear();
      infile.seekg(counter.at(frame), ios::beg);
      cstring.clear();
      
      while (cstring != "CRYS" && !infile.eof()) {
         getline(infile, cline);
         //cout << cline << endl;
         cstring.assign(cline, 0, 4);
         if (cstring == "ATOM" || cstring == "HETA") {
            // Clear cjunk
	         for (int i = 0; i < 80; i++) {
	            cjunk[i] = ' ';
            }
        	   //      COLUMNS      DATA TYPE        FIELD      DEFINITION
	         // ------------------------------------------------------
	         //  1 -  6      Record name      "ATOM    "
	         //  7 - 11      Integer          serial     Atom serial number.
	         // 13 - 16      Atom             name       Atom name.
	         // 17           Character        altLoc     Alternate location indicator.
	         // 18 - 20      Residue name     resName    Residue name.
	         // 22           Character        chainID    Chain identifier.
	         // 23 - 26      Integer          resSeq     Residue sequence number.
	         // 27           AChar            iCode      Code for insertion of residues
	         // 31 - 38      Real(8.3)        x          Orthogonal coordinates for X 
	         //                                          in  Angstroms
	         // 39 - 46      Real(8.3)        y          Orthogonal coordinates for Y
	         //                                          in Angstroms
	         // 47 - 54      Real(8.3)        z          Orthogonal coordinates for Z 
	         //                                          in Angstroms
	         // 55 - 60      Real(6.2)        occupancy  Occupancy.
	         // 61 - 66      Real(6.2)        tempFactor Temperature factor.
	         // 77 - 78      LString(2)       element    Element symbol
	         // 79 - 80      LString(2)       charge     Charge on the atom.
	         
	         for (int i = 0; i < 6; i++) {
	            cjunk[i] = (char) cline.at(i+6);
   	      }
	         atno = atoi(cjunk);
            //cout << atno << endl;
	         
	         type.at(count) = cline.at(13);
	         
	         for (int i = 0; i < 3; i++) {
	            for (int j = i*8+31; j < i*8+38; j++) { 
	               cjunk[j-30-i*8] = (char) cline.at(j);
               }
               coord.at(i) = atof(cjunk);
	         }
	         
	         // Get ANISOU line
	         getline(infile, cline);
	         // Check it's really the correct line
	         cstring.assign(cline, 0, 4);	 
	         if (cstring != "ANIS") {
	            cout << cline << endl;
	            cout << "Something's not right. This should be a ANIOSU line." << endl;
	            cout << "But it's not." << endl;
               cout << "Press Enter to exit.";
               cin.clear();
               cin.ignore(1,0);
               std::exit(0);
            }
	         //cout << cline << endl;
            // PDB format:
	         // COLUMNS     DATA TYPE         FIELD          DEFINITION
	         // --------------------------------------------------------
	         //  1 - 6      Record name       "ANISOU"
	         //  7 - 11     Integer           serial         Atom serial number.
	         // 13 - 16     Atom              name           Atom name.
	         // 17          Character         altLoc         Alternate location indicator
	         // 18 - 20     Residue name      resName        Residue name.
	         // 22          Character         chainID        Chain identifier.
	         // 23 - 26     Integer           resSeq         Residue sequence number.
	         // 27          AChar             iCode          Insertion code.
	         // 29 - 35     Integer           u[0][0]        U(1,1)
	         // 36 - 42     Integer           u[1][1]        U(2,2)
	         // 43 - 49     Integer           u[2][2]        U(3,3)
	         // 50 - 56     Integer           u[0][1]        U(1,2)
	         // 57 - 63     Integer           u[0][2]        U(1,3)
	         // 64 - 70     Integer           u[1][2]        U(2,3)
	         // 77 - 78     LString(2)        element        Element symbol, right-justified.
	         // 79 - 80     LString(2)        charge         Charge on the atom.       
            
            // Clear cjunk
	         for (int i = 0; i < 80; i++) {
	            cjunk[i] = ' ';
            }
	         
	         for (int i = 0; i < 6; i++) {
	            for (int j = i*7+29; j < i*7+35; j++) { 
	               cjunk[j-29-i*7] = (char) cline.at(j);
               }
               aniso.at(i) = atof(cjunk);
	         }
	         
	         // Create ANISOU matrix
	         gsl_matrix_set(A, 0, 0, aniso.at(0)*1e-4);
	         gsl_matrix_set(A, 1, 1, aniso.at(1)*1e-4);
	         gsl_matrix_set(A, 2, 2, aniso.at(2)*1e-4);
	         gsl_matrix_set(A, 0, 1, aniso.at(3)*1e-4);
	         gsl_matrix_set(A, 0, 2, aniso.at(4)*1e-4);
	         gsl_matrix_set(A, 1, 2, aniso.at(5)*1e-4);
	         
	         gsl_matrix_set(A, 1, 0, aniso.at(3)*1e-4);
	         gsl_matrix_set(A, 2, 0, aniso.at(4)*1e-4);
	         gsl_matrix_set(A, 2, 1, aniso.at(5)*1e-4);
            
	         // Let's use a temporary matrix for our calculations, 
	         // so we can test the results with A when we're done
	         info = gsl_matrix_memcpy(mtemp, A);
	         info = gsl_eigen_symmv(mtemp, eval, evec, w);
	         
            // Eigenvectors are stored as columns in evec
	         for (int i = 0; i < n; i++) {
	            gsl_vector_set(vec0, i, gsl_matrix_get(evec, i, 0));
            }
	         for (int i = 0; i < n; i++) {
	            gsl_vector_set(vec1, i, gsl_matrix_get(evec, i, 1));
            }
	         for (int i = 0; i < n; i++) {
	            gsl_vector_set(vec2, i, gsl_matrix_get(evec, i, 2));
            }
            
            // Calculate quaternions from the eigenvalues
	         // Q = (W, X, Y, Z)
	         // trc(P) = 1 + P00 + P11 + P22
	         // trc(P) > 0, else something wrong
	         // S = 1/(2*sqrt(trc((P))))
	         // W = 1/(4*S)
	         // X = (P21 - P12) * S
	         // Y = (P02 - P20) * S
	         // Z = (P10 - P01) * S
	         // The eigenvalues are the corresponding half-radii
	         
	         trcP = 1 + gsl_matrix_get(evec,0,0) +  gsl_matrix_get(evec,1,1) +  gsl_matrix_get(evec,2,2);
	         
            if (trcP > 0.0) {
	            W = 0.5*sqrt(trcP);
	            X = (gsl_matrix_get(evec,2,1) - gsl_matrix_get(evec,1,2)) / (4.0*W);
	            Y = (gsl_matrix_get(evec,0,2) - gsl_matrix_get(evec,2,0)) / (4.0*W);
	            Z = (gsl_matrix_get(evec,1,0) - gsl_matrix_get(evec,0,1)) / (4.0*W);
	            norm = sqrt(W*W + X*X + Y*Y + Z*Z);
            }
	         else {
               m00 = gsl_matrix_get(evec,0,0);
      	      m01 = gsl_matrix_get(evec,0,1);
      	      m02 = gsl_matrix_get(evec,0,2);
      	      m10 = gsl_matrix_get(evec,1,0);
      	      m11 = gsl_matrix_get(evec,1,1);
      	      m12 = gsl_matrix_get(evec,1,2);
      	      m20 = gsl_matrix_get(evec,2,0);
      	      m21 = gsl_matrix_get(evec,2,1);
      	      m22 = gsl_matrix_get(evec,2,2);
               
	            if ((m00 > m11) & (m00 > m22)) { 
	               S = sqrt( 1.0 + m00 - m11 - m22 ) * 2; // S=4*qx 
	               W = (m21 - m12) / S;
	               X = 0.25 * S;
	               Y = (m01 + m10) / S; 
	               Z = (m02 + m20) / S; 
               } 
	            else if (m11 > m22) { 
	               S = sqrt( 1.0 + m11 - m00 - m22 ) * 2; // S=4*qy
	               W = (m02 - m20) / S;
	               X = (m01 + m10) / S; 
	               Y = 0.25 * S;
	               Z = (m12 + m21) / S; 
               } 
               else { 
	               S = sqrt( 1.0 + m22 - m00 - m11 ) * 2; // S=4*qz
	               W = (m10 - m01) / S;
	               X = (m02 + m20) / S; 
	               Y = (m12 + m21) / S; 
	               Z = 0.25 * S;
               }
            }
            
	         // Write to the dumpfile and the infile
            //cout << count+1 << " " << count+1 << " " << 
            //   coord.at(0) << " " << coord.at(1) << " "<< coord.at(2) << " " << W << " " << X << " " << Y << " " << Z << endl;
            //infile << gsl_vector_get(eval, 0) << " " << gsl_vector_get(eval, 1) << " " << gsl_vector_get(eval, 2) << "          " << type.at(count) << endl;
	         
	         
	         // Insert geometry data in the geom vector.
	         geom.push_back(count+1);
	         geom.push_back(count+1);
	         geom.push_back(coord.at(0));
	         geom.push_back(coord.at(1));
	         geom.push_back(coord.at(2));
	         geom.push_back(W);
	         geom.push_back(X);
	         geom.push_back(Y);
	         geom.push_back(Z);
	         
	         // Insert ellipsoid radii data in the shape vector.
	         shape.push_back(gsl_vector_get(eval, 0));
	         shape.push_back(gsl_vector_get(eval, 1));
	         shape.push_back(gsl_vector_get(eval, 2));
	         
	         // Create appropriate color for the atom type
	         //cout << "Type: " << type.at(count) << endl;
	         if (type.at(count) == 'C') {
	            color.push_back(0.0);
	            color.push_back(1.0);
	            color.push_back(0.0);
            }
            else if (type.at(count) == 'N') {
               color.push_back(1.0);
	            color.push_back(0.0);
	            color.push_back(0.0);
            }
            else if (type.at(count) == 'O') {
               color.push_back(0.0);
	            color.push_back(0.0);
	            color.push_back(1.0);
            }
            else if (type.at(count) == 'H') {
               color.push_back(1.0);
	            color.push_back(1.0);
	            color.push_back(1.0);
            }
            else if (type.at(count) == 'S') {
               color.push_back(0.5);
	            color.push_back(0.5);
	            color.push_back(0.0);
            }
	         else {
               color.push_back(1.0);
	            color.push_back(1.0);
	            color.push_back(0.0);
            }
             
	         // Create bond vector for the atom
	         bondvec.push_back(0.0);
	         bondvec.push_back(0.0);
	         bondvec.push_back(0.0);
            // increase counter	 
	         count++;
	         if (count >= (int)type.size()) {
	            type.resize(2*count);
            }
         } // Done everything that's needs to be done for the ATOM and ANISOU keywords
      }  // Reached end of file
      //species = count;
      //atoms = count;
      //cout << "The number of species is " << species << endl;
      //showTS = false;
      
      //for (int i = 0; i < count; i++) {
      //   cout << geom.at(i*9+0) << " " << geom.at(i*9+1) << " " <<
      //           geom.at(i*9+2) << " " << geom.at(i*9+3) << " " << geom.at(i*9+4) << " " << geom.at(i*9+5) << 
      //           geom.at(i*9+6) << " " << geom.at(i*9+7) << " " << geom.at(i*9+8) << " " << endl;
      //}
      //for (int i = 0; i < count; i++) {
      //   cout << type.at(i) << " " << color.at(i*3+0) << " " << color.at(i*3+1) << " " << color.at(i*3+2) << " " << endl;
      //}
      
   }
   else {
      cout << "Something is seriously wrong. The input file seems to be closed." << endl;
      cout << "Press Enter to exit.";
      cin.clear();
      cin.ignore(1,0);
      std::exit(0);
   }
}


/* DOSTUFF ***************************************************************** */
void DoStuff() {

   int type, no;
   vector<double> trans(3), quat(4), mean(3);
   
   // Get the average position in all three directions
   mean = average();
   
   for (int i = 0; i < atoms ; i++) {
      no       = (int) geom.at(i*9+0);
      type     = (int) geom.at(i*9+1);
      trans.at(0) =       geom.at(i*9+2) - mean.at(0); // geom.at(c*9+2)
      trans.at(1) =       geom.at(i*9+3) - mean.at(1); // geom.at(c*9+3)
      trans.at(2) =       geom.at(i*9+4) - mean.at(2); // geom.at(c*9+4)
      quat.at(0)  =       geom.at(i*9+5);
      quat.at(1)  =       geom.at(i*9+6);
      quat.at(2)  =       geom.at(i*9+7);
      quat.at(3)  =       geom.at(i*9+8);
      CreateEllipsoid(type, trans, quat, 30, 30);
   }
   glPushMatrix();
   if (solidBonds) {
      for (int i = 0; i < (int) topology.size()/4; i++) {
         DrawSolidBonds(i, mean);
      }
   }
   else {
      for (int i = 0; i < (int) topology.size()/4; i++) {
         DrawLineBonds(i, mean);
      }
   }
   glPopMatrix();
}


/* CREATEELLIPSOID ********************************************************* */
void CreateEllipsoid(int type, vector<double> trans, vector<double> quat, int s1, int s2) {

   double Ax, Ay, Az, alpha;
   double s, x, y, z, w;
   double rx, ry, rz;
   // material properties (refl. coeff.)
   float white_rc[] = {0.8, 0.8, 0.8, 1},
         spec_rc[]  = {0.1, 0.1, 0.1, 1}, //{1.0, 1.0 ,1.0 ,1.0}, //  
         color_rc[] = {color.at((type-1)*3+0), color.at((type-1)*3+1), color.at((type-1)*3+2), 1},
         spec_exp   = 0; //100 ; //
   
   // define material properties for front face
   glMaterialfv(GL_FRONT, GL_AMBIENT,   white_rc);
   glMaterialfv(GL_FRONT, GL_DIFFUSE,   color_rc);
   glMaterialfv(GL_FRONT, GL_SPECULAR,  spec_rc);
   glMaterialf (GL_FRONT, GL_SHININESS, spec_exp);
   
   // calculate rotation axis and angle from quaternions
   w = quat.at(0);
   x = quat.at(1);
   y = quat.at(2);
   z = quat.at(3);
   s = sqrt(x*x + y*y + z*z);
   // Avoid divide by zero
   if (s <= 0)
      s = 1;
      
   alpha = 2.0*acos(w)*180.0/pi;
   Ax = x/s;
   Ay = y/s;
   Az = z/s;
   
   rx = shape.at((type-1)*3+0);
   ry = shape.at((type-1)*3+1);
   rz = shape.at((type-1)*3+2);
   
   // Spheres get special treatment
   if (rx == ry && ry == rz) {
      s1 = (int) s1/2; // fewer polygons
      s2 = (int) s2/2;  
   }
   
   glPushMatrix();
   glTranslatef(trans.at(0), trans.at(1), trans.at(2));
   glRotatef(alpha, Ax, Ay, Az);
   glScalef(rx, ry, rz);
   
   gluQuadricNormals(q, GL_SMOOTH);	// Generate Smooth Normals For The Quad
   gluQuadricTexture(q, GL_TRUE);	// Enable Texture Coords For The Quad
   glColor3f(1.0f, 1.0f, 1.0f);
   
   //glEnable(GL_TEXTURE_2D);
   glBindTexture(GL_TEXTURE_2D, texture.at(0));
   
   // create the polygons and normals for a sphere;
   // radius, resolution along longitudes and latitudes
   // rotate and scale sphere to create ellipsoid

   gluSphere(q, 1, s1, s2);
   glPopMatrix();
}


/* DRAWSOLIDBONDS ********************************************************** */
void DrawSolidBonds(int i, vector<double> mean) {

   // Now with quatratics to create cylindrical bonds!
   //double p1[3], p2[3], color[] = {0.1, 0.21, 0.31}; // Dark grey
   double p1[3], p2[3]; //, mid[3];  
   float white_rc[] = {0.8, 0.8, 0.8, 1},
         spec_rc[]  = {0.1, 0.1, 0.1, 1}, //{1.0, 1.0 ,1.0 ,1.0}, //  
         color_rc[] = {0.81, 0.71, 0.71},
         spec_exp   = 0;
   double x1, y1, z1, x2, y2, z2, w, x, y, z, len, xylen;
   int type1, type2, it;
   double phi, theta;
   
   if((int) geom.size() > 0) { 
      // Calculate begining of bond
      type1 = (int) geom.at((topology.at(i*4)-1)*9 + 1);
      x1 =          geom.at((topology.at(i*4)-1)*9 + 2) - mean.at(0);
      y1 =          geom.at((topology.at(i*4)-1)*9 + 3) - mean.at(1);
      z1 =          geom.at((topology.at(i*4)-1)*9 + 4) - mean.at(2);
      
      w =           geom.at((topology.at(i*4)-1)*9 + 5);
      x =           geom.at((topology.at(i*4)-1)*9 + 6);
      y =           geom.at((topology.at(i*4)-1)*9 + 7);
      z =           geom.at((topology.at(i*4)-1)*9 + 8);
      
      it = (topology.at(i*4+2)-1)*species*3 + (type1-1)*3;
      p1[0] = -(1-2*y*y-2*z*z)*bondvec.at(it+0) + 
              -(2*x*y-2*w*z)  *bondvec.at(it+1) + 
              -(2*x*z+2*w*y)  *bondvec.at(it+2) + x1;
      p1[1] = -(2*x*y+2*w*z)  *bondvec.at(it+0) + 
              -(1-2*x*x-2*z*z)*bondvec.at(it+1) + 
              -(2*y*z-2*w*x)  *bondvec.at(it+2) + y1;
      p1[2] = -(2*x*z-2*w*y)  *bondvec.at(it+0) + 
              -(2*y*z+2*w*x)  *bondvec.at(it+1) +
              -(1-2*x*x-2*y*y)*bondvec.at(it+2) + z1;

      //Calculate end of bond
      type2 = (int) geom.at((topology.at(i*4+1)-1)*9 + 1);
      x2 =          geom.at((topology.at(i*4+1)-1)*9 + 2) - mean.at(0);
      y2 =          geom.at((topology.at(i*4+1)-1)*9 + 3) - mean.at(1);
      z2 =          geom.at((topology.at(i*4+1)-1)*9 + 4) - mean.at(2);
    
      w =           geom.at((topology.at(i*4+1)-1)*9 + 5);
      x =           geom.at((topology.at(i*4+1)-1)*9 + 6);
      y =           geom.at((topology.at(i*4+1)-1)*9 + 7);
      z =           geom.at((topology.at(i*4+1)-1)*9 + 8);
      
      it = (topology.at(i*4+3)-1)*species*3 + (type2-1)*3;
      p2[0] = -(1-2*y*y-2*z*z)*bondvec.at(it+0) + 
              -(2*x*y-2*w*z)  *bondvec.at(it+1) + 
              -(2*x*z+2*w*y)  *bondvec.at(it+2) + x2;
      p2[1] = -(2*x*y+2*w*z)  *bondvec.at(it+0) + 
              -(1-2*x*x-2*z*z)*bondvec.at(it+1) + 
              -(2*y*z-2*w*x)  *bondvec.at(it+2) + y2;
      p2[2] = -(2*x*z-2*w*y)  *bondvec.at(it+0) + 
              -(2*y*z+2*w*x)  *bondvec.at(it+1) +
              -(1-2*x*x-2*y*y)*bondvec.at(it+2) + z2;
      
      // Calculate length of bond
      len = 0;
      for (int j = 0; j < 3; j++) {
         len = len + (p2[j]-p1[j])*(p2[j]-p1[j]);
      }
      len = sqrt(len);
      
      xylen = 0;
      for (int j = 0; j < 2; j++) {
         xylen = xylen + (p2[j]-p1[j])*(p2[j]-p1[j]);
      }
      xylen = sqrt(xylen);
      
      //cout << "Len: " << len << endl;
      
      glPushMatrix();
      //glDisable(GL_CULL_FACE);         // We need to draw both sides of the bond
      gluQuadricNormals(q, GL_SMOOTH);	// Generate Smooth Normals For The Quad
      gluQuadricTexture(q, GL_FALSE);	// Enable Texture Coords For The Quad
      glColor3f(1.0f, 1.0f, 1.0f);
      glMaterialfv(GL_FRONT, GL_AMBIENT,   white_rc);
      glMaterialfv(GL_FRONT, GL_DIFFUSE,   color_rc);
      glMaterialfv(GL_FRONT, GL_SPECULAR,  spec_rc);
      glMaterialf (GL_FRONT, GL_SHININESS, spec_exp);
      
      // Center The Cylinder
      glTranslatef(p1[0], p1[1], p1[2]);
      
      double sign;
      if (fabs((p2[1]-p1[1])) > 0.0)
         sign = fabs(p2[1]-p1[1])/(p2[1]-p1[1]);
      else
         sign = 1.0;
      
      phi   = sign * acos((p2[0]-p1[0])/xylen) * 180.0f/pi;  // angle against X
      theta = acos((p2[2]-p1[2])/len) * 180.0f/pi; // angle against Z
      
      glRotatef(phi, 0, 0, 1);    // Rotate phi around Z 
      glRotatef(theta, 0, 1, 0 ); // Rotate theta around Y
      
      //glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
      gluQuadricNormals(q, GL_SMOOTH);	// Generate Smooth Normals For The Quad
      
      if (texture.size() == 2) {
         // We have bond texture!
         gluQuadricTexture(q, GL_TRUE);
         glBindTexture(GL_TEXTURE_2D, texture.at(1));
      }
      gluCylinder(q, bondwidth, bondwidth, len, 5, 5);
      //glEnable(GL_CULL_FACE);
      //glCullFace(GL_BACK);
      //glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
      glPopMatrix();
   }
}


/* DRAWLINEBONDS *********************************************************** */
void DrawLineBonds(int i, vector<double> mean) {

   // Now with quatratics to create cylindrical bonds!
   //double p1[3], p2[3], color[] = {0.1, 0.21, 0.31}; // Dark grey
   float p1[3], p2[3]; //, mid[3];  
   float color_rc[] = {0.61, 0.51, 0.51};
   double x1, y1, z1, x2, y2, z2, w, x, y, z;
   int type1, type2, it;
   
   if((int) geom.size() > 0) { 
      // Calculate begining of bond
      type1 = (int) geom.at((topology.at(i*4)-1)*9 + 1);
      x1 =          geom.at((topology.at(i*4)-1)*9 + 2) - mean.at(0);
      y1 =          geom.at((topology.at(i*4)-1)*9 + 3) - mean.at(1);
      z1 =          geom.at((topology.at(i*4)-1)*9 + 4) - mean.at(2);
      
      w =           geom.at((topology.at(i*4)-1)*9 + 5);
      x =           geom.at((topology.at(i*4)-1)*9 + 6);
      y =           geom.at((topology.at(i*4)-1)*9 + 7);
      z =           geom.at((topology.at(i*4)-1)*9 + 8);
      
      it = (topology.at(i*4+2)-1)*species*3 + (type1-1)*3;
      p1[0] = -(1-2*y*y-2*z*z)*bondvec.at(it+0) + 
              -(2*x*y-2*w*z)  *bondvec.at(it+1) + 
              -(2*x*z+2*w*y)  *bondvec.at(it+2) + x1;
      p1[1] = -(2*x*y+2*w*z)  *bondvec.at(it+0) + 
              -(1-2*x*x-2*z*z)*bondvec.at(it+1) + 
              -(2*y*z-2*w*x)  *bondvec.at(it+2) + y1;
      p1[2] = -(2*x*z-2*w*y)  *bondvec.at(it+0) + 
              -(2*y*z+2*w*x)  *bondvec.at(it+1) +
              -(1-2*x*x-2*y*y)*bondvec.at(it+2) + z1;

      //Calculate end of bond
      type2 = (int) geom.at((topology.at(i*4+1)-1)*9 + 1);
      x2 =          geom.at((topology.at(i*4+1)-1)*9 + 2) - mean.at(0);
      y2 =          geom.at((topology.at(i*4+1)-1)*9 + 3) - mean.at(1);
      z2 =          geom.at((topology.at(i*4+1)-1)*9 + 4) - mean.at(2);
    
      w =           geom.at((topology.at(i*4+1)-1)*9 + 5);
      x =           geom.at((topology.at(i*4+1)-1)*9 + 6);
      y =           geom.at((topology.at(i*4+1)-1)*9 + 7);
      z =           geom.at((topology.at(i*4+1)-1)*9 + 8);
      
      it = (topology.at(i*4+3)-1)*species*3 + (type2-1)*3;
      p2[0] = -(1-2*y*y-2*z*z)*bondvec.at(it+0) + 
              -(2*x*y-2*w*z)  *bondvec.at(it+1) + 
              -(2*x*z+2*w*y)  *bondvec.at(it+2) + x2;
      p2[1] = -(2*x*y+2*w*z)  *bondvec.at(it+0) + 
              -(1-2*x*x-2*z*z)*bondvec.at(it+1) + 
              -(2*y*z-2*w*x)  *bondvec.at(it+2) + y2;
      p2[2] = -(2*x*z-2*w*y)  *bondvec.at(it+0) + 
              -(2*y*z+2*w*x)  *bondvec.at(it+1) +
              -(1-2*x*x-2*y*y)*bondvec.at(it+2) + z2;
      
      glMaterialfv(GL_FRONT, GL_DIFFUSE, color_rc);
      glLineWidth((GLdouble) 1.8);
      glColor3fv(color_rc);
      //glPushMatrix();
      
      glBegin(GL_LINES);
      glVertex3fv(p2); // start line at ellipsoid centered at {x2,y2,z2}
      glVertex3fv(p1); // end line at ellipsoid centered at {x1,y1,z1}
      glEnd();

        
   }
}


/* PRINTIMAGE ************************************************************** */
void PrintImage(bool enter) {

   ILuint ilimage;
   ILboolean success = false;
   int i;
   char junk[50];
   string fname;
   
   if (enter) {
      fname = "Screen_";
      if (printcount < 1000) fname+= "0"; // Add leading zero
      if (printcount < 100) fname+= "0"; // Add leading zero
      if (printcount < 10)  fname+= "0"; // Add leading zero
      i = sprintf(junk, "%d", printcount);
      fname+= junk;
      printcount++;     
   }
   else {           
      fname = inname;
   }
   fname += "_"; // Add underscore
   if (frame < 1000) fname+= "0"; // Add leading zero
   if (frame < 100) fname+= "0"; // Add leading zero
   if (frame < 10)  fname+= "0"; // Add leading zero
   i = sprintf(junk, "%d", frame);
   fname+= junk;
   fname+= format;
   
   if (print) {
      ilEnable(IL_FILE_OVERWRITE);
      ilGenImages(1, &ilimage);
      ilBindImage(ilimage);
      ilutGLScreen();
      //success = ilSaveImage((wchar_t *) fname.c_str());
      success = ilSaveImage((char *) fname.c_str());
      ilDeleteImages(1, &ilimage);
   }
   if (enter) {
      if (success) {
         cout << "Success! File " << fname << " has been written." << endl;
      }
   }
   if (!success) {
      cout << "Fail! No file of name " << fname << " can be written." << endl;
      cout << "I don't know why. Do you have write permissions?" << endl;
      cout << "Press Enter / Return to exit.";
      cin.clear();
      cin.ignore(1,0);
      std::exit(0);
   }
}


/* CREATEGLUTMENUS ********************************************************* */
void createGLUTMenus() {

   int lightmenu, printmenu, tsmenu, menu, formatmenu;
   
   lightmenu = glutCreateMenu(MenuHandler);
   glutAddMenuEntry("Stationary Light", 1);
   glutAddMenuEntry("Moving Light", 2);
   
   printmenu = glutCreateMenu(MenuHandler);
   glutAddMenuEntry("Print Images", 3);
   glutAddMenuEntry("Do Not Print Images", 4);
   glutAddMenuEntry("Print This Screen (Enter)", 5);
   
   tsmenu = glutCreateMenu(MenuHandler);
   glutAddMenuEntry("Hide Timestep (t)", 6);
   glutAddMenuEntry("Show Timestep", 7);
   
   formatmenu = glutCreateMenu(MenuHandler);
   glutAddMenuEntry("JPG", 8);
   glutAddMenuEntry("PNG", 9);
   glutAddMenuEntry("BMP", 10);
   
   menu = glutCreateMenu(MenuHandler);
   glutAddSubMenu("Light", lightmenu);
   glutAddSubMenu("Print", printmenu);
   glutAddSubMenu("File format", formatmenu);
   glutAddSubMenu("Timestep", tsmenu);
   glutAddMenuEntry("Quit (Q)", 0);
   
   glutAttachMenu(GLUT_RIGHT_BUTTON);
}


/* MENUHANDLER ************************************************************* */
void MenuHandler(int id) {

   string fname;
   //char junk[50];
   //int i;
   
   switch (id)
   {
      case 0:
         CloseFile();
         cout << "Bye." << endl;
         std::exit(0); // Quit
         break;
      case 1:
         rotating_light = 0;
         cout << "Stationary light" << endl;
	      break;
	   case 2: 
         rotating_light = 1;
         cout << "Rotating light" << endl;
	      break;
      case 3: 
         print = true;
         break;
      case 4: 
         print = false;
	      break;
      case 5: 
         print = true;
         PrintImage(true);
         //printcount++;
         print = false;
	      break;
      case 6: 
         showTS = false;
	      break;
      case 7: 
         showTS = true;
	      break;
      case 8:
         format = ".jpg";
	      break;
      case 9:
         format = ".png";
         break;
      case 10:
         format = ".bmp";
	      break;
   }
}


/* SHOWTIMESTEP ************************************************************* */
void ShowTimeStep() {

   int i;
   char junk[256];
   float color[] = {2.0, 2.0, 2.0}; //{0.71, 0.61, 0.61};
   
   // We don't want light, depth or textures  
   glDisable(GL_LIGHT0);
   glDisable(GL_LIGHT1);
   glDisable(GL_TEXTURE_2D);
   glDisable(GL_DEPTH_TEST);
   
   setOrthographicProjection();
   
   glPushMatrix();
   glLoadIdentity();
   i = sprintf(junk, "Timestep %i of %i", frame, (int)counter.size()-1);
   glMaterialfv(GL_FRONT, GL_AMBIENT, color);
   glColor3fv(color);
   renderBitmapString(20.0, window_height-20.0, (void *)bitfont, junk);
   glPopMatrix();
   
   resetPerspectiveProjection();
   
   glEnable(GL_DEPTH_TEST);
   
   // Reenable textures
   glEnable(GL_TEXTURE_2D);
   // Reenable light
   glEnable(GL_LIGHT0);
   glEnable(GL_LIGHT1);
}


/* SETORTHOGRAPHICPROJECTION ************************************************ */
void setOrthographicProjection() {

	// switch to projection mode
	glMatrixMode(GL_PROJECTION);
	// save previous matrix which contains the 
	//settings for the perspective projection
	glPushMatrix();
	// reset matrix
	glLoadIdentity();
	// set a 2D orthographic projection
	gluOrtho2D(0, window_width, 0, window_height);
	// invert the y axis, down is positive
	glScalef(1, -1, 1);
	// mover the origin from the bottom left corner
	// to the upper left corner
	glTranslatef(0, (GLdouble)-window_height, 0);
	glMatrixMode(GL_MODELVIEW);
}


/* RESETPERSPECTIVEPROJECTION *********************************************** */
void resetPerspectiveProjection() {

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}


/* RENDERBITMAPSTRING ******************************************************* */
void renderBitmapString(double x, double y, void *bmfont, char *instring) {

   char *c;
   glRasterPos2f(x, y);
   for (c=instring; *c != '\0'; c++) {
      glutBitmapCharacter(bmfont, *c);
   }
}


/* RESHAPE ***************************************************************** */
void Reshape(int width, int height) { // new size in pixels 

    if (height == 0)
       height = 1;
    
    glViewport(0, 0, width, height);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f, (GLdouble)width/(GLdouble)height, 1, window_depth);
    
    glMatrixMode(GL_MODELVIEW);
    
    window_width  = width;
    window_height = height;
}


/* KEYHANDLER ************************************************************** */
void KeyHandler(unsigned char key, int x, int y) {

   double ex[3], ey[3], ez[3],
         zaxis[3] = {0, 0, 1},
         rotAxis[3] = {0, 0, 0};
   string bondfile, conffile, st;
   //char junk[50];
   //int i;
   
   if (GLUT_KEY_DOWN) {
      /* Quit */
      if ((key == 'Q') | (key == 27)) { // Quit on Esc
         CloseFile();
         cout << "Bye." << endl;
         std::exit(0);
	      return;
      }
      /* Movement */
      if (key == '+' || key == '=') { // Zoom in
         getMVCoord(ex, ey, ez);
         glTranslatef(2.0*dot(ex,zaxis), 2.0*dot(ey,zaxis), 2.0*dot(ez,zaxis));
         //eyex = (1-0.04)*eyex + focusx*0.04;
         //eyey = (1-0.04)*eyey + focusy*0.04;
         //eyez = (1-0.04)*eyez + focusz*0.04;
         //glLoadIdentity();
         //gluLookAt(eyex, eyex, eyez, focusx, focusy, focusz, 0,1,0);
      }
      if (key == '-' || key == '_') { // Zoom out
         getMVCoord(ex, ey, ez);
         glTranslatef(-2.0*dot(ex,zaxis), -2.0*dot(ey,zaxis), -2.0*dot(ez,zaxis));
         //eyex = (1+0.04)*eyex - focusx*0.04;
         //eyey = (1+0.04)*eyey - focusy*0.04;
         //eyez = (1+0.04)*eyez - focusz*0.04;
         //glLoadIdentity();
         //gluLookAt(eyex, eyex, eyez, focusx, focusy, focusz, 0,1,0);
      }
      if (key == 'd') { // Rotate around window y-axis
         rotAxis[1] = 1;
         getMVCoord(ex, ey, ez);
         glRotatef(2, dot(rotAxis, ex), dot(rotAxis, ey), dot(rotAxis, ez));
      }
      if (key == 'a') { // Rotate around negative window y-axis
         rotAxis[1] = -1;
         getMVCoord(ex, ey, ez);
         glRotatef(2, dot(rotAxis, ex), dot(rotAxis, ey), dot(rotAxis, ez));
      }
      if (key == 'w') { // Rotate around window x-axis
         rotAxis[0] = -1;
         getMVCoord(ex, ey, ez);
         glRotatef(2, dot(rotAxis, ex), dot(rotAxis, ey), dot(rotAxis, ez));
      }
      if (key == 's') { // Rotate around negative window x-axis
         rotAxis[0] = 1;
         getMVCoord(ex, ey, ez);
         glRotatef(2, dot(rotAxis, ex), dot(rotAxis, ey), dot(rotAxis, ez));
      }
      if (key == 'd') { // Rotate around window x-axis
         rotAxis[1] = 1;
         getMVCoord(ex, ey, ez);
         glRotatef(2, dot(rotAxis, ex), dot(rotAxis, ey), dot(rotAxis, ez));
      }
      if (key == 'a') { // Rotate around negative window x-axis
         rotAxis[1] = -1;
         getMVCoord(ex, ey, ez);
         glRotatef(2, dot(rotAxis, ex), dot(rotAxis, ey), dot(rotAxis, ez));
      }
      if (key == 'q') { // Rotate around window negative z-axis
         rotAxis[2] = 1;
         getMVCoord(ex, ey, ez);
         glRotatef(2, dot(rotAxis, ex), dot(rotAxis, ey), dot(rotAxis, ez));
      }
      if (key == 'e') { // Rotate around window z-axis
         rotAxis[2] = -1;
         getMVCoord(ex, ey, ez);
         glRotatef(2, dot(rotAxis, ex), dot(rotAxis, ey), dot(rotAxis, ez));
      }
      /* Animation */
      if (key == 32) { // Space pauses animation
         if (animate) animate = false;
         else animate = true;
         //cout << "Space pressed. Animate = " << animate << endl;
      }
      /* Go to frame */
      if (key == 'p') {
         if (frame > 0) {
            //cout << "p pressed. frame = " << frame << endl;
            frame--;
            //cout << "frame decreased by one. frame = " << frame << endl;
            infile.clear();
            infile.seekg(counter.at(frame), ios::beg);
            //cout << "counter.at(" << frame << ") = " << counter.at(frame) << endl << ", " << infile.tellg() << endl;
            ReadFrame(); 
            animate = false;
            //cout << "p: frame = " << frame << endl;
         }
      }
      if (key == 'n') { 
         if (frame < (int) counter.size()-1) { 
            frame++;
            //cout << "n: frame = " << frame << endl;
            infile.clear();
            infile.seekg(counter.at(frame), ios::beg);
            //cout << "counter.at(" << frame << ") = " << counter.at(frame) << endl << ", " << infile.tellg() << endl;
            ReadFrame(); 
            animate = false;
            //cout << "n: frame = " << frame << endl;
         }
      }
      if (key == 'P') {
         if ((frame - (int) counter.size()/10)> 0) {
            //cout << "p pressed. frame = " << frame << endl;
            frame -= (int) counter.size()/10;
            //cout << "frame decreased by one. frame = " << frame << endl;
            infile.clear();
            infile.seekg(counter.at(frame), ios::beg);
            //cout << "counter.at(" << frame << ") = " << counter.at(frame) << endl << ", " << infile.tellg() << endl;
            ReadFrame(); 
            animate = false;
            //cout << "p: frame = " << frame << endl;
         }
         else {
            frame = 0;
            infile.clear();
            infile.seekg(counter.at(frame), ios::beg);
            //cout << "counter.at(" << frame << ") = " << counter.at(frame) << endl << ", " << infile.tellg() << endl;
            ReadFrame(); 
            animate = false;
         }
      }
      if (key == 'N') { 
         if ((frame + (int) counter.size()/10) < (int) counter.size()-1) { 
            frame += (int) counter.size()/10;
            //cout << "n: frame = " << frame << endl;
            infile.clear();
            infile.seekg(counter.at(frame), ios::beg);
            //cout << "counter.at(" << frame << ") = " << counter.at(frame) << endl << ", " << infile.tellg() << endl;
            ReadFrame(); 
            animate = false;
            //cout << "n: frame = " << frame << endl;
         }
         else {
            frame = counter.size()-1;
            infile.clear();
            infile.seekg(counter.at(frame), ios::beg);
            ReadFrame(); 
            animate = false;
         }
      }
      if (key == 'G') {
         char Gstring[80]; 
         int newframe;
         cout << "Go to frame: ";
         cin >> Gstring;
         if (!isdigit(Gstring[0])) {
            cin.clear();
            cout << "I'm sorry Dave, I'm afraid I can't do that." << endl;
         }
         else {
            newframe = atoi(Gstring);
            if ((int) newframe < 0) {
               cout << "The frame number must be larger than zero!" << endl;
               frame = 0;
            }
            else if ((int) newframe > (int) counter.size()-1) {
               cout << "Only " << (int) counter.size()-1 << " frames in this file." << endl;
               frame = counter.size()-1;
            }
            else frame = newframe;
            
            cout << "Going to frame " << frame << endl;  
            cin.clear();
            infile.clear();
            infile.seekg(counter.at(frame), ios::beg);
            //cout << "counter.at(" << frame << ") = " << counter.at(frame) << endl << ", " << infile.tellg() << endl;
            ReadFrame();
         }
         animate = false;
      }
      if (key == 'r') { 
         // Going back to the first frame of the file
         frame = 0;
         //cout << "r: frame = " << frame << endl;
         //cout << "'r' pressed" << endl;
         infile.clear(); // Any errors from EOF needs to be cleared
         infile.seekg(counter.at(frame), ios::beg);
         ReadFrame();
         animate = false;
         //cout << "Frame read" << endl;
      }
      if (key == 'f') {
         // Going to the last frame of the file
         frame = counter.size()-1;
         //cout << "Frame " << frame << endl;
         infile.clear();
         infile.seekg(counter.at(frame), ios::beg);
         ReadFrame();
         animate = false;
      }
      /* Load new file */
      if (key == 'L') { 
         print = false;
         frame = 0;
         CloseFile();
         infile.clear(); // Clean file buffer errors
         cin.clear(); // Clean cin buffer
         counter.clear();
         DeleteTextures();
         
         cout << "Please enter name of file with configuration information: " ;
         getline (cin, conffile);
         title = "Biovec - " + conffile;
         
         bondfile = ReadConfig(conffile);
         glutSetWindowTitle(title.c_str());
         cout << "Config file read." << endl; 
         
         if (ellipsoidtexture.length() != 0)
            LoadTextures(ellipsoidtexture, 0);
         if (bondtexture.length() != 0)
            LoadTextures(bondtexture, 1);
         
         glClearColor(BGColor[0], BGColor[1], BGColor[2], BGColor[3]);
         glFogfv(GL_FOG_COLOR, FogColor);
         
         OpenFile();
         infile.seekg(0, ios::beg);
         ReadFrame();
         
         topology.clear();
         if (!bondfile.empty()) {
            ReadBonds(bondfile);
         }
         
         animate = false;
         cout << "Done. Rendering." << endl;
      }
      /* Fog */
      if (key == 'g')	{
         if (gp) {
            glDisable(GL_FOG);
            gp = false;
         }
         else {
            glEnable(GL_FOG);
            gp = true;
         }
      }
      /* Time step */ 
      if (key == 't') {
         if (showTS)
            showTS = false;
         else
            showTS = true;
      }
      /* Print image */
      if (key == 13) { // Enter
         print = true;
         PrintImage(true);
         //printcount++;
         print = false;
      }
      /* Any other key */
      else
         return;
   }
   
   // the position of a light is affected by M, so...
   if (rotating_light) { // Transform by M
      glLightfv(GL_LIGHT0, GL_POSITION, light_pos_0);
      glLightfv(GL_LIGHT1, GL_POSITION, light_pos_1);
   } 
   else {              // Stationary light
      glPushMatrix();
      glLoadIdentity();
      glLightfv(GL_LIGHT0, GL_POSITION, light_pos_0);
      glLightfv(GL_LIGHT1, GL_POSITION, light_pos_1);
      glPopMatrix();
   }
   glutPostRedisplay(); // update image
}


/* SPECIALKEY ************************************************************* */
void SpecialKey(int key, int x, int y) {

   double ex[3], ey[3], ez[3],
         xaxis[] = {1, 0, 0},
         yaxis[] = {0, 1, 0};
   
   switch(key) {
      case GLUT_KEY_RIGHT:
         // Calculate ModelView coorinates, translate
         getMVCoord(ex, ey, ez);
         glTranslatef(dot(ex,xaxis), dot(ey,xaxis), dot(ez,xaxis));
         break;
      case GLUT_KEY_LEFT:
         getMVCoord(ex, ey, ez);
         glTranslatef(-dot(ex,xaxis), -dot(ey,xaxis), -dot(ez,xaxis));
         break;
      case GLUT_KEY_DOWN:
         getMVCoord(ex, ey, ez);
         glTranslatef(-dot(ex,yaxis), -dot(ey,yaxis), -dot(ez,yaxis));
         break;
      case GLUT_KEY_UP:
         getMVCoord(ex, ey, ez);
         glTranslatef(dot(ex,yaxis), dot(ey,yaxis), dot(ez,yaxis));
         break;
   }
   if (rotating_light) { // Transform by M
      glLightfv(GL_LIGHT0, GL_POSITION, light_pos_0);
      glLightfv(GL_LIGHT1, GL_POSITION, light_pos_1);
   } 
   else {              // Stationary light
     glPushMatrix();
     glLoadIdentity();
     glLightfv(GL_LIGHT0, GL_POSITION, light_pos_0);
     glLightfv(GL_LIGHT1, GL_POSITION, light_pos_1);
     glPopMatrix();
   }
   glutPostRedisplay(); // update image
}


/* GETMVCOORD ************************************************************** */
void getMVCoord(double *ex, double *ey, double *ez) {

   float MV[16];
   int i;
   
   glGetFloatv(GL_MODELVIEW_MATRIX, MV);
   for (i = 0; i < 3; ++i) {
      ex[i] = MV[i];
      ey[i] = MV[i+4];
      ez[i] = MV[i+8];
   }
}


/* DOT ********************************************************************* */
double dot(double *a, double *b) {

   return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
}


/* AVERAGE ***************************************************************** */
vector<double> average() {

   vector<double> c(3);
   
   for (int i = 0; i < atoms; i++) {
      c.at(0) +=     geom.at(i*9+2);
      c.at(1) +=     geom.at(i*9+3);
      c.at(2) +=     geom.at(i*9+4);
   }
   for (int i = 0; i < 3; i++) {
      c.at(i) /= atoms;
   }
   
   return(c);
}


/* LOADTEXTURES ************************************************************ */
void LoadTextures(string fname, int i) {

   ILuint texid;
   ILboolean success;
   
   ilGenImages(1, &texid); // Generation of one image name 
   ilBindImage(texid);     // Binding of image name 
   
   //success = ilLoadImage((wchar_t *) fname.c_str());
   success = ilLoadImage((char *) fname.c_str());
   
   if (success) {
      //cout << "Texture " << fname << " loaded." << endl;
      texture.at(i) = ilutGLBindMipmaps ();
      if (texture.at(i) != 0)
         cout << "Texture " << fname << " bound." << endl;
      else
         cout << "ilutOglBuildMipmaps() did not work." << endl;
   }
   else {
      cout << "ilLoadImage(" << fname << ") did not work." << endl;
   }
   ilDeleteImages(1, &texid); 
}


/* DELETETEXTURES ********************************************************** */
void DeleteTextures() {

   for (int i = 0; i < (int) texture.size(); i++) {
      glDeleteTextures(1, &texture.at(i));
   }
}


/* HANDLEMOUSESTATE ******************************************************** */
void HandleMouseState(int button, int state, int x, int y) {

   //int key;
   
   // update our button state
   //cout << "Button " << button << " pressed" << endl; 
   if (button == 0) {
      if (state == GLUT_DOWN) {
         MouseState.leftButton = true;
         MouseState.func_key = glutGetModifiers();
      }
      else
         MouseState.leftButton = false;
   }
   if (button == GLUT_RIGHT_BUTTON) {
	   if (state == GLUT_DOWN)
         MouseState.middleButton = true;
      else
		   MouseState.middleButton = false;
   }
	if(button == GLUT_MIDDLE_BUTTON) {
 	   if(state == GLUT_DOWN)
		   MouseState.rightButton = true;
	   else
		   MouseState.rightButton = false;
   }
   
   // update our position so we know a delta when the mouse is moved
   MouseState.mx = x;
   MouseState.my = y;
   
   //PrintMouseState();
}


/* PRINTMOUSESTATE ********************************************************* */
void PrintMouseState() {

   printf("mouse x: %d, y: %d left: %s, middle: %s, right: %s,\n", MouseState.mx,
		MouseState.my, MouseState.leftButton ? "down":"up", 
      MouseState.middleButton ? "down":"up", MouseState.rightButton ? "down":"up");
}


/* HANDLEMOUSEMOVE ********************************************************* */
void HandleMouseMove(int x, int y) {

   double ex[3], ey[3], ez[3],
         zaxis[3] = {0, 0, 1},
         rotAxis[3] = {0, 0, 0};
   
   // calculate a delta in movement
   int yDelta = MouseState.my - y;
   int xDelta = MouseState.mx - x;
   
   // commit the mouse position
   MouseState.mx = x;
   MouseState.my = y;
   
   //specialKey = glutGetModifiers();
   // when we need to rotate (only the left button is down)
   if (MouseState.leftButton) {
      if (MouseState.func_key == GLUT_ACTIVE_SHIFT) {
         // rotate delta about Z axis
         rotAxis[2] = -1;
         getMVCoord(ex, ey, ez);
         glRotatef(-xDelta, dot(rotAxis, ex), dot(rotAxis, ey), dot(rotAxis, ez));
      }
      else if (MouseState.func_key == GLUT_ACTIVE_CTRL ) {
           // zoom
           getMVCoord(ex, ey, ez);
           //if (abs(yDelta) > 0) {
              //glTranslatef(2.0*dot(ex,zaxis), 2.0*dot(ey,zaxis), 2.0*dot(ez,zaxis));
              glTranslatef(yDelta*dot(ex,zaxis), 
                 yDelta*dot(ey,zaxis), 
                 yDelta*dot(ez,zaxis));
           //}
      }
      else {
         // rotate delta about X and Y axis
         rotAxis[1] = 1;
         getMVCoord(ex, ey, ez);
         glRotatef(-xDelta, dot(rotAxis, ex), dot(rotAxis, ey), dot(rotAxis, ez));
         rotAxis[0] = 1;
         rotAxis[1] = 0;
         getMVCoord(ex, ey, ez);
         glRotatef(-yDelta, dot(rotAxis, ex), dot(rotAxis, ey), dot(rotAxis, ez));
      }
   }
   
   //PrintMouseState();
   
   // the position of a light is affected by M, so...
   if (rotating_light) { // Transform by M
      glLightfv(GL_LIGHT0, GL_POSITION, light_pos_0);
      glLightfv(GL_LIGHT1, GL_POSITION, light_pos_1);
   } 
   else {              // Stationary light
      glPushMatrix();
      glLoadIdentity();
      glLightfv(GL_LIGHT0, GL_POSITION, light_pos_0);
      glLightfv(GL_LIGHT1, GL_POSITION, light_pos_1);
      glPopMatrix();
   }
   glutPostRedisplay(); // update image
}


/* HANDLEWHEEL ************************************************************* */
void HandleWheel(int wheel, int dir, int x, int y) {

   double ex[3], ey[3], ez[3],
   zaxis[3] = {0, 0, 1};
   //rotAxis[3] = {0, 0, 0};
   
   getMVCoord(ex, ey, ez);
   
   // The direction of the wheel is contrary to what I want... Therefore -dir
   glTranslatef(-dir*2.0*dot(ex,zaxis), -dir*2.0*dot(ey,zaxis), -dir*2.0*dot(ez,zaxis));
}



